module;

#include <windows.h>

export module pige.example;

export import pige.winapi.msglog;
import std;

export namespace pige::examples
{
	class queue_window : public pige::winapi::logging_window
	{
	public:
		using logging_window::logging_window;

	protected:
		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			switch (msg)
			{
				case WM_KEYDOWN:
					return on_keydown(wparam);
				case WM_USER+1:
					pige::winapi::logger::log("Ping!");
					return 5;
				default:
					return window::window_proc(msg, wparam, lparam);
			}
		}

	private:
		LRESULT on_keydown(WPARAM keyCode)
		{
			switch (keyCode)
			{
				case 'S':
				{
					fall_asleep();
					break;
				}
				case 'P':
					InvalidateRect(*this, nullptr, true);
					break;
				case 'C':
					DestroyWindow(*this);
					break;
				case 'T':
					SetTimer(*this, 1, 200, nullptr);
					break;
				case 'U':
					RedrawWindow(*this, nullptr, nullptr, RDW_INVALIDATE | RDW_ERASE | RDW_UPDATENOW | RDW_ERASENOW);
					break;
				case 'M':
					message_thread();
					break;

			}
			return 0;
		}

		void fall_asleep()
		{
			using namespace std::chrono_literals;
			int len = GetWindowTextLengthW(*this);
			std::wstring text(len, L'\0');
			GetWindowTextW(*this, text.data(), len);
			SetWindowTextW(*this, L"ZzzZzz");
			std::this_thread::sleep_for(2s);
			SetWindowTextW(*this, text.c_str());
		}

		void message_thread()
		{
			std::thread t(send_messages, this);
			t.detach();
		}

		static void send_messages(queue_window *w)
		{
			using namespace std::chrono_literals;
			int n = 3;
			while (n--)
			{
				std::this_thread::sleep_for(1s);
				//Messages sent from a separate thread will still be handled by the thread that created the window
				//GetMessageW dispatches them to window procedure directly, they do not pass through our message loop.
				auto r = SendMessageW(*w, WM_USER + 1, 0, 0);
				OutputDebugStringW(std::format(L"Message response: {}\n", r).c_str());

				//Try posting messages instead, see that they are returned by GetMessageW.
				//No way to recive response though.
				//PostMessageW(*w, WM_USER + 1, 0, 0);
			}
		}
	};
}