module;

#include <windows.h>

export module pige.example;

export import pige.winapi;
import std;

export namespace pige::examples
{
	class autotype_window : public pige::winapi::window
	{
	public:
		autotype_window(HINSTANCE instance, LPCWSTR title)
			: window{ instance, pige::winapi::window_desc{ title, { 300, 200 } } }
		{
			if (!RegisterHotKey(*this, 1, MOD_CONTROL | MOD_SHIFT, L'T'))
				throw winapi::winapi_error{};
		}

	protected:
		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			switch (msg)
			{
				case WM_KEYDOWN:
					return on_keydown(wparam);
				case WM_TIMER:
					return on_timer();
				case WM_HOTKEY:
					SetTimer(*this, 1, 500, nullptr);
				case WM_DESTROY:
					UnregisterHotKey(*this, 1);
				default:
					return window::window_proc(msg, wparam, lparam);
			}
		}

	private:
		LRESULT on_keydown(WPARAM keyCode)
		{
			switch (keyCode)
			{
				case 'T':
					SetTimer(*this, 1, 500, nullptr);
					ShowWindow(*this, SW_MINIMIZE);
					break;
			}
			return 0;
		}

		LRESULT on_timer()
		{
			using namespace std::chrono_literals;
			static size_t char_idx = 0;
			//constexpr wchar_t text[] = L"👻";
			const KEYBDINPUT keystrokes[][4] = {
				{ { VK_SHIFT }, { 'B' }, { 'B', 0, KEYEVENTF_KEYUP }, { VK_SHIFT, 0, KEYEVENTF_KEYUP } },
				{ { 'O' }, { 'O', 0, KEYEVENTF_KEYUP } },
				{ { 'O' }, { 'O', 0, KEYEVENTF_KEYUP } },
				{ { VK_SHIFT },	{ '1' }, { '1', 0, KEYEVENTF_KEYUP } },
				{ { '1' }, { '1', 0, KEYEVENTF_KEYUP },	{ VK_SHIFT, 0, KEYEVENTF_KEYUP } },
				{ { VK_SPACE }, { VK_SPACE, 0, KEYEVENTF_KEYUP } },
				//To get the values either use text[0] and text[1] or find an UTF-16 encoder and convert 👻.
				//The characters typed by the previous elements could be entered using KEYEVENTF_UNICODE
				//instead of composing them from simulated virtual key presses.
				{ { 0, L'\xD83D', KEYEVENTF_UNICODE }, { 0, L'\xDC7B', KEYEVENTF_UNICODE } },
				{ { VK_CONTROL }, { VK_ADD }, { VK_ADD, 0, KEYEVENTF_KEYUP } },
				{ { VK_ADD }, { VK_ADD, 0, KEYEVENTF_KEYUP }, { VK_CONTROL, 0, KEYEVENTF_KEYUP } },
				{ { VK_RWIN }, { VK_UP }, { VK_UP, 0, KEYEVENTF_KEYUP }, { VK_RWIN, 0, KEYEVENTF_KEYUP } },
				{ { VK_CONTROL }, { VK_DIVIDE }, { VK_DIVIDE, 0, KEYEVENTF_KEYUP }, { VK_CONTROL, 0, KEYEVENTF_KEYUP } },
				{ { VK_MENU }, { VK_F4 }, { VK_F4, 0, KEYEVENTF_KEYUP }, { VK_MENU, 0, KEYEVENTF_KEYUP } }
			};
			constexpr size_t count = ARRAYSIZE(keystrokes);
			for (int key_idx = 0; key_idx < 4; ++key_idx)
			{
				if (!keystrokes[char_idx][key_idx].wVk && !keystrokes[char_idx][key_idx].wScan)
					break;
				INPUT input{ .type = INPUT_KEYBOARD, .ki = keystrokes[char_idx][key_idx] };
				SendInput(1, &input, sizeof(INPUT));
			}
			++char_idx;
			if (char_idx >= count)
			{
				KillTimer(*this, 1);
				std::this_thread::sleep_for(500ms);
				ShowWindow(*this, SW_RESTORE);
				char_idx = 0;
			}
			return 0;
		}
	};
}