module;

#include <windows.h>
#include <windowsx.h>

export module pige.winapi.msglog:format.convert;

import :format.common_types;
import :format.name_maps;

/*
* Struct initializers and convertes
*/

export
{
	//Initializes rectangle with (x,y) as top-left corner coordinates and (cx,cy) as size
	RECT toRect(int x, int y, int cx, int cy)
	{
		return { x, y, x + cx, y + cy };
	}

	//Initializes rectangle with specified top-left corner coordinates p and size s
	RECT toRect(POINT p, SIZE s)
	{
		return toRect(p.x, p.y, s.cx, s.cy);
	}

	//Initializes rectangle with specified top-left corner coordinates p and size s
	RECT toRect(POINT p, POINT s)
	{
		return toRect(p.x, p.y, s.x, s.y);
	}

	SIZE toSize(POINT p)
	{
		return { p.x, p.y };
	}

	/*
	 * LPARAM converters
	 */

	 //Extracts x and y coordinates from LPARAM
	POINT toCoords(LPARAM lParam)
	{
		return { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };
	}

	//Returns referenct to RECT lParam is pointing to
	RECT &toRect(LPARAM lParam)
	{
		return *reinterpret_cast<RECT *>(lParam);
	}

	//Extracts width and height from lParam
	SIZE toSize(LPARAM lParam)
	{
		return { LOWORD(lParam), HIWORD(lParam) };
	}

	//Returns reference to WINDOWPOS lParam is pointing to
	WINDOWPOS &toWndPos(LPARAM lParam)
	{
		return *reinterpret_cast<WINDOWPOS *>(lParam);
	}

	//Returns referenct to NCCALCSIZE_PARAMS lParam is pointing to
	NCCALCSIZE_PARAMS &toNCCSP(LPARAM lParam)
	{
		return *reinterpret_cast<NCCALCSIZE_PARAMS *>(lParam);
	}

	//Returns reference to MINMAXINFOR lParam is pointing to
	MINMAXINFO &toMMI(LPARAM lParam)
	{
		return *reinterpret_cast<MINMAXINFO *>(lParam);
	}

	//Returns reference to CREATESTRUCTW lParam is pointing to
	CREATESTRUCTW &toCSW(LPARAM lParam)
	{
		return *reinterpret_cast<CREATESTRUCTW *>(lParam);
	}

	//Extracts key repeat count from lParam
	UINT toKeyRepeatCount(LPARAM lParam)
	{
		return static_cast<UINT>(lParam & 0x7F);
	}

	//Extracts previous key state (true is down, false if up) from lParam
	bool toPrevKeyState(LPARAM lParam)
	{
		return ((lParam >> 30) & 0x1) != 0;
	}

	//Extracts window handle from lParam
	HWND toHWnd(LPARAM lParam)
	{
		return reinterpret_cast<HWND>(lParam);
	}

	//Extracts hit test result from lParam
	hit_test toHitTest(LPARAM lParam)
	{
		return hit_test{ LOWORD(lParam) };
	}

	//Extracts mouse message type from lParam
	UINT toMouseMsg(LPARAM lParam)
	{
		return HIWORD(lParam);
	}

	//Returns pointer to character buffer lParam is poitning to
	wchar_t *toCharBuf(LPARAM lParam)
	{
		return reinterpret_cast<wchar_t *>(lParam);
	}

	mk_keystate toMKKeystate(LPARAM lParam)
	{
		return mk_keystate{ GET_KEYSTATE_LPARAM(lParam) };
	}

	app_cmd toAppCmd(LPARAM lParam)
	{
		return app_cmd{ GET_APPCOMMAND_LPARAM(lParam) };
	}

	/*
	 * WPARAM convertes
	 */

	 //Extracts BOOL value from wParam
	BOOL toBool(WPARAM wParam)
	{
		return static_cast<BOOL>(wParam);
	}

	//Extracts XButton index from wParam
	UINT toXButton(WPARAM wParam)
	{
		return HIWORD(wParam);
	}

	//Extracts virtual key code from wParam
	vk_code toVKCode(WPARAM wParam)
	{
		return vk_code{ static_cast<USHORT>(wParam) };
	}

	//Extract character code form wParam
	wchar_t toChar(WPARAM wParam)
	{
		return static_cast<wchar_t>(wParam);
	}

	//Extracts wheel delta from wParam
	UINT toWheelDistance(WPARAM wParam)
	{
		return HIWORD(wParam);
	}

	//Extracts window handle from wParam
	HWND toHWnd(WPARAM wParam)
	{
		return reinterpret_cast<HWND>(wParam);
	}

	//Extracts child message type from wParam
	UINT toChildMsg(WPARAM wParam)
	{
		return LOWORD(wParam);
	}

	//Extracts child id from wParam
	UINT toChildID(WPARAM wParam)
	{
		return HIWORD(wParam);
	}

	//Extracts system command type from wParam
	sys_cmd toSysCmd(WPARAM wParam)
	{
		UINT sc = GET_SC_WPARAM(wParam);
		return sys_cmd{ sc };
	}

	//Extreacts scroll bar request form wParam
	USHORT toSBRequest(WPARAM wParam)
	{
		return LOWORD(wParam);
	}

	//Extracts scroll bar position from wParam
	USHORT toSBPos(WPARAM wParam)
	{
		return HIWORD(wParam);
	}

	mk_keystate toMKKeystate(WPARAM wParam)
	{
		return mk_keystate{ LOWORD(wParam) };
	}

	hit_test toHitTest(WPARAM wParam)
	{
		return hit_test{ GET_NCHITTEST_WPARAM(wParam) };
	}

	/*
	 * LRESULT converters
	 */
	hit_test toHTResult(LRESULT r)
	{
		return hit_test{ r };
	}

	/*
	 * Helper output operators for various structs
	 */

	 //Prints coordinates of a point
	std::wostream &operator<<(std::wostream &o, const POINT &p)
	{
		return o << L"[x:" << p.x << L",y:" << p.y << L']';
	}

	//Prints width and height of a size
	std::wostream &operator<<(std::wostream &o, const SIZE &s)
	{
		return o << L"[w:" << s.cx << L",h:" << s.cy << L']';
	}

	//Prints top-left corner coordinates and size of a rectangle
	std::wostream &operator<<(std::wostream &o, const RECT &r)
	{
		return o << L"[x:" << r.left << L",y:" << r.top <<
			L",w:" << (r.right - r.left) << L",h:" << (r.bottom - r.top) << L']';
	}

	std::wostream &operator<<(std::wostream &o, const HWND &hWnd)
	{
		//Use cache to limit WM_GETTEXTLENGTH and WM_GETTEXT messages sent
		auto it = wndMap.find(hWnd);
		if (it != wndMap.end())
		{
			//If string is empty, this is a recursive call, print pointer
			if (it->second.empty())
				return o << reinterpret_cast<void *>(hWnd);
			return o << it->second;
		}
		//Temporary place empty string to block recursive calls
		wndMap.emplace(hWnd, L"");
		size_t len = 0;
		std::wstring name;
		if (GetCurrentThreadId() == GetWindowThreadProcessId(hWnd, nullptr))
		{
			//If our own thread, call DefWindowProcW directly
			//Text passed as window name on creation is fine enough (it's just logging),
			//so we don't care about custom WM_GETTEXT handling.
			len = DefWindowProcW(hWnd, WM_GETTEXTLENGTH, 0, 0);
			if (len)
			{
				name.resize(len + 1);
				DefWindowProcW(hWnd, WM_GETTEXT, len + 1, reinterpret_cast<LPARAM>(name.data()));
			}
		}
		else
		{
			len = GetWindowTextLengthW(hWnd);
			if (len)
			{
				name.resize(len + 1);
				GetWindowTextW(hWnd, name.data(), static_cast<int>(len + 1));
			}
		}
		if (len == 0)
		{
			//No text associated with window (might be before WM_NCCREATE is processed. Print the handle
			//and class name if possible.
			wndMap.erase(hWnd);
			o << reinterpret_cast<void *>(hWnd);
			//Try to get window class name
			std::wstring cls_name(255, L'\0');
			len = GetClassNameW(hWnd, cls_name.data(), 256);
			if (len)
			{
				cls_name.resize(len);
				o << L'(' << cls_name << L')';
			}
			return o;

		}
		//Store text in cache and print it
		name.resize(len);
		wndMap[hWnd] = name;
		return o << L'"' << name << L'"';
	}

	//Prints size, position, z-order and flags of WINDOWPOS
	std::wostream &operator<<(std::wostream &o, const WINDOWPOS &p)
	{
		o << toRect(p.x, p.y, p.cx, p.cy) << L",after:" << p.hwndInsertAfter;
		if (p.flags)
		{
			LPCWSTR flags[] = { L"noSize", L"noMove", L"noZOrder", L"noRedraw", L"noActivate", L"drawFrame", L"show", L"hide" };
			bool comma = false;
			o << L" [";
			for (int i = 0; i < 8; ++i)
			{
				if (p.flags >> i & 0x1)
				{
					if (comma)
						o << L',';
					else
						comma = true;
					o << flags[i];
				}
			}
			o << L']';
		}
		return o;
	}

	//Print state of keys for mouse messages
	std::wostream &operator<<(std::wostream &o, mk_keystate s)
	{
		o << L'[';
		auto us = get(s);
		if (us)
		{
			LPCWSTR buttons[] = { L"l", L"r", L"shift", L"ctrl", L"m", L"x1", L"x2" };
			bool comma = false;
			for (int i = 0; i < 7; ++i)
			{
				if (us >> i & 0x1)
				{
					if (comma)
						o << L',';
					else
						comma = true;
					o << buttons[i];
				}
			}
		}
		return o << L']';
	}

	//Prints rectangle of maximized window and minimum and maximum size of restored window
	std::wostream &operator<<(std::wostream &o, const MINMAXINFO &inf)
	{
		return o << L"maximized:" << toRect(inf.ptMaxPosition, inf.ptMaxSize) << L",minSize:" <<
			toSize(inf.ptMinTrackSize) << L",maxSize:" << toSize(inf.ptMaxTrackSize);
	}

	//Prints name of the created window stored in CREATESTRUCTW
	std::wostream &operator<<(std::wostream &o, const CREATESTRUCTW &csw)
	{
		return o << L"window:\"" << csw.lpszName << "\"";
	}

	std::wostream &operator<<(std::wostream &o, vk_code c)
	{
		auto it = vkMap.find(c);
		if (it != vkMap.end())
			o << it->second;
		else
		{
			if (std::iswupper(get(c)))
			{
				o << static_cast<wchar_t>(get(c));
			}
			else
			{
				auto flags = o.flags();;
				o << std::hex << std::showbase << get(c);
				o.flags(flags);
			}
		}
		return o;
	}

	std::wostream &operator<<(std::wostream &o, hit_test r)
	{
		auto it = htMap.find(r);
		if (it != htMap.end())
			o << it->second;
		else
			o << get(r);
		return o;
	}

	std::wostream &operator<<(std::wostream &o, sys_cmd s)
	{

		auto it = scMap.find(s);
		if (it != scMap.end())
			o << it->second;
		else
		{
			auto flags = o.flags();;
			o << std::hex << std::showbase << get(s);
			o.flags(flags);
		}
		return o;
	}

	std::wostream &operator << (std::wostream &o, app_cmd a)
	{
		auto it = acMap.find(a);
		if (it != acMap.end())
			o << it->second;
		else
		{
			auto flags = o.flags();;
			o << std::hex << std::showbase << get(a);
			o.flags(flags);
		}
		return o;
	}
}