module;

#include <windows.h>
#include <windowsx.h>

export module pige.winapi.msglog:window;

export import pige.winapi;
export import :logger;
export import :format;

export namespace pige::winapi
{
	class logging_window : public window
	{
	public:
		static constexpr std::wstring_view default_class_name{ L"Message Logging Class" };
		inline static class_desc default_class(HINSTANCE instance)
		{
			class_desc desc{ instance, default_class_name.data() };
			desc.lpfnWndProc = window_proc;
			return desc;
		}

		logging_window() = default;
		logging_window(logging_window &&other) = default;
		
		inline logging_window(HINSTANCE instance, LPCWSTR title)
			: logging_window{ instance, window_desc{ title } }
		{ }

	protected:
		inline logging_window(HINSTANCE instance, window_desc const &desc)
			: logging_window{ default_class(instance), desc }
		{ }
		logging_window(class_desc const &cls_desc, window_desc const &desc)
			: window{ cls_desc, desc }
		{ }

		using window::window_proc;

		static LRESULT CALLBACK window_proc(HWND handle, UINT msg, WPARAM wparam, LPARAM lparam)
		{
			logging_window *w = nullptr;
			try
			{
				w = reinterpret_cast<logging_window *>(msg == WM_NCCREATE ?
					handle_created(handle, *reinterpret_cast<LPCREATESTRUCTW>(lparam)) :
					get_window(handle));

				MSG m = make_msg(handle, msg, wparam, lparam);
				logger::start_section(L"++window_proc processing", m);
				LRESULT r;
				if (w)
				{
					r = w->window_proc(msg, wparam, lparam);
					if (msg == WM_NCDESTROY)
						handle_destroyed(w);
				}
				else r = DefWindowProcW(handle, msg, wparam, lparam);
				logger::end_section(L"--window_proc handled", std::make_pair(m, r));
				return r;
			}
			catch (winapi_error &e)
			{
				logger::log(e.what());
			}
			catch (std::exception &e)
			{
				logger::log(e.what());
			}
			catch (...)
			{
				logger::log(L"Unknown error!");
			}
			if (msg == WM_NCDESTROY)
				handle_destroyed(w);
			PostQuitMessage(0);
			return 0;
		}

	private:
		[[nodiscard]] static MSG make_msg(HWND handle, UINT type, WPARAM wparam, LPARAM lparam)
		{
			DWORD pos = GetMessagePos();
			return { handle, type, wparam, lparam, static_cast<DWORD>(GetMessageTime()),
					 { GET_X_LPARAM(pos), GET_Y_LPARAM(pos) } };
		}
	};
}