export module pige.utility;

import std;

namespace pige::utility
{
	export template <class Tag, typename T>
	struct strong_typedef
	{
		explicit constexpr operator T &() & noexcept
		{
			return value;
		}

		explicit constexpr operator const T &() const &noexcept
		{
			return value;
		}

		explicit constexpr operator T && () && noexcept
		{
			return std::move(value);
		}

		explicit constexpr operator const T && () const &&noexcept
		{
			return std::move(value);
		}

		[[nodiscard]] bool operator ==(strong_typedef const &) const noexcept = default;

		T value;
	};

	//not exported
	template <class Tag, typename T>
	T to_underlying_type(strong_typedef<Tag, T>);

	export template <class StrongTypedef>
	using underlying_type
		= decltype(to_underlying_type(std::declval<typename std::decay<StrongTypedef>::type>()));

	export template <class Tag, typename T>
	constexpr T &get(strong_typedef<Tag, T> &type) noexcept
	{
		return static_cast<T &>(type);
	}

	export template <class Tag, typename T>
	constexpr const T &get(const strong_typedef<Tag, T> &type) noexcept
	{
		return static_cast<const T &>(type);
	}

	export template <class Tag, typename T>
	constexpr T &&get(strong_typedef<Tag, T> &&type) noexcept
	{
		return static_cast<T &&>(static_cast<T &>(type));
	}

	export template <class Tag, typename T>
	constexpr const T &&get(const strong_typedef<Tag, T> &&type) noexcept
	{
		return static_cast<const T &&>(static_cast<const T &>(type));
	}

	export template <class StrongTypedef>
	struct hashable : std::hash<underlying_type<StrongTypedef>>
	{
		using underlying_type = underlying_type<StrongTypedef>;
		using underlying_hash = std::hash<underlying_type>;

		std::size_t operator()(const StrongTypedef &lhs) const
			noexcept(noexcept(underlying_hash{}(std::declval<underlying_type>())))
		{
			return underlying_hash{}(static_cast<const underlying_type &>(lhs));
		}
	};
}