module;

#include <windows.h>

export module pige.winapi:application;

export import :window;
import std;

namespace pige::winapi
{
	//Just like with the window class, any modification to the program's behaviour should be done in a sub-class
	export class application
	{
		HINSTANCE m_instance;
	public:
		application(HINSTANCE instance)
			: m_instance{ instance }
		{ }

		[[nodiscard]] virtual int run(int show_command)
		{
			window w{ m_instance, L"Main Window" };
			return run(show_command, w);
		}

		[[nodiscard]] virtual int run(int show_command, window &main_window)
		{
			main_window.m_is_main = true;
			ShowWindow(main_window, show_command);
			return main_loop();
		}
		
	protected:
		inline HINSTANCE instance() const { return m_instance; }

		[[nodiscard]] virtual int main_loop()
		{
			MSG msg{};
			BOOL result;
			while (0 != (result = GetMessageW(&msg, nullptr, 0, 0)))
			{
				if (result < 0)
					throw winapi::winapi_error{};

				TranslateMessage(&msg);
				DispatchMessageW(&msg);
			}
			return static_cast<int>(msg.wParam);
		}
	};

	//Functions that simplify starting the application from main, providing exception handling around instatiation
	//of the application class and the call to run method.

#define TRY try
#define CATCH_ALL								\
	catch (winapi_error &e)						\
	{											\
		OutputDebugStringW(e.what().c_str());	\
	}											\
	catch (std::exception &e)					\
	{											\
		OutputDebugStringA(e.what());			\
	}											\
	catch (...)									\
	{											\
		OutputDebugStringW(L"Unknown Error!");	\
	}											\
	return 1;									

	export template<typename TApp = application>
	[[nodiscard]] int start_app(HINSTANCE instance, int show_command) noexcept
	{
		TRY {
			TApp app{ instance };
			return app.run(show_command);
		} CATCH_ALL
	}

	export template<typename TWindow, typename TApp = application>
	[[nodiscard]] int start_app(HINSTANCE instance, int show_command, LPCWSTR title)
	{
		TRY{
			TApp app{ instance };
			TWindow window{ instance, title };
			return app.run(show_command, window);
		} CATCH_ALL
	}

#undef TRY
#undef CATCH_ALL
}