module;

#include <windows.h>

export module pige.examples:timer;

import pige.winapi;
import std;

export namespace pige::examples
{
	class animation_timer : public pige::winapi::window
	{
	public:
		animation_timer(HINSTANCE instance, LPCWSTR title)
			: window{ instance, title }
		{
			//Box should be moved every 50ms
			SetTimer(*this, 1, 50, nullptr);
		}

	protected:
		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			switch (msg)
			{
				case WM_TIMER:
					on_timer();
					return 0;

				//DefaultWindowProc will erase the whole update region with the window class's background brush
				//If you want to erase it yourself (or, in this example, prentend to) return non-zero for this message
				//This will cause the box to leave a trail behind, even if you tell the window to erase background
				//in on_timer
				/*case WM_ERASEBKGND:
					return 1;*/

				case WM_PAINT:
				{
					on_paint();
					return 0;
				}
				default:
					return window::window_proc(msg, wparam, lparam);
			}
		}

	private:
		//Initial position
		RECT m_box = { 348, 112, 448, 212 };
		//Distance traveled by the box in each step
		SIZE m_vel = { 17, 13 };

		void on_timer()
		{
			RECT rc_client;
			GetClientRect(*this, &rc_client);
			OffsetRect(&m_box, m_vel.cx, m_vel.cy);

			//Calculate how far the box lies beyond client area
			LONG dx = 0, dy = 0;
			if (m_box.left < 0)
				dx = m_box.left; //negative
			else if (m_box.right > rc_client.right)
				dx = (m_box.right - rc_client.right); //positive
			if (m_box.top < 0)
				dy = m_box.top;
			else if (m_box.bottom > rc_client.bottom)
				dy = (m_box.bottom - rc_client.bottom);

			//Bounce the rectangle - however far it was hanging outside the client area, is how far it should be moved
			//from the client area bounds towards the centre
			OffsetRect(&m_box, -2*dx, -2*dy);

			//Flip the movement direction if we bounced of an edge
			if (dx != 0)
				m_vel.cx *= -1;
			if (dy != 0)
				m_vel.cy *= -1;

			//Force the entire window to repaint
			InvalidateRect(*this, nullptr,
				true /*passing false will not erase the window's background, leaveing a trail behind the square*/);
			//We could be more selective and only invalidate the area conaining the previous and the new box positions:
			//store initial m_box in a temporary, and after new position is calculated, call UnionRect, and pass
			//as the second argument

			//Effect of InvalidateRect can be achieved with RedrawWindow, e.g.:
			//RedrawWindow(*this, nullptr, nullptr, RDW_INVALIDATE | RDW_ERASE);
			
			//to redraw immediately instead of waiting for WM_PAINT to show up in the message queue, add:
			//UpdateWindow(*this);
			
			//or instead of the Invalidate/Update combo:
			//RedrawWindow(*this, nullptr, nullptr, RDW_INVALIDATE | RDW_ERASE | RDW_UPDATENOW | RDW_ERASENOW);
		}

		void on_paint()
		{
			PAINTSTRUCT p{ };
			//Drawing context is needed to draw on the window:
			auto context = BeginPaint(*this, &p);

			//Draw a box:
			FillRect(context, &m_box, reinterpret_cast<HBRUSH>(COLOR_ACTIVECAPTION + 1));

			//Clean up:
			EndPaint(*this, &p);

			//The unfortunate side effect of the background being erased is that animation will flicker. That would
			//still happen if we erased background ourselves here instead of relying on the DefaultWindowProc.
			//Next example shows how to deal with that.
		}
	};
}