module;

#include <windows.h>
#include <windowsx.h>
#include <cassert>

export module pige.examples;

import std;
export import pige.winapi;

export namespace pige::examples
{
	class sniper : public pige::winapi::window
	{
	public:
		sniper(HINSTANCE instance, LPCWSTR title)
			: window{ instance, winapi::window_desc{ title, { 300, 200 } } }
		{ }

	protected:
		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			switch (msg)
			{
				case WM_LBUTTONDOWN:
					//Capture the mouse, so we get mouse events outside our window,
					//instead of the window below the cursor
					SetCapture(*this);
					return 0;
				case WM_LBUTTONUP:
					//Capture should generaly be released as soon as possible, and the OS will take it away from us
					//if we are to gready.
					ReleaseCapture();

					//Many older sources use HIWORD/LOWORD macros to extract position form lparam, but that will give
					//incorrect results, e.g. with negative coordinates which happen if the cursor is on a screen
					//to the left of the primary display on multi-monitor setups.
					snipe({ GET_X_LPARAM(lparam), GET_Y_LPARAM(lparam) });
					return 0;
				default:
					return window::window_proc(msg, wparam, lparam);
			}
		}
	private:

		void snipe(POINT p)
		{
			//convert from client coordinates to screen coordinates
			ClientToScreen(*this, &p);

			//snipe a window
			auto handle = WindowFromPoint(p);
			if (!handle) return;

			//find its root top-level window
			handle = GetAncestor(handle, GA_ROOT);
			if (!handle) return;

			//print the window
			print_window(handle, L"Children of: ");

			//Children hierarchy enumerated in a depth-first order, so to figure out the depth of a window in the
			//hierarchy tree, we need just to keep the depth and handle of the previously examined window. When the next
			//one is enumerated we can adjust the depth by searching for the common parent of it and the previous one
			std::pair<HWND, size_t> prev_depth{ handle, 0 };

			//print the children
			EnumChildWindows(handle, enum_proc, reinterpret_cast<LPARAM>(&prev_depth));
		}

		static void print_window(HWND handle, LPCWSTR prefix)
		{
			//Get the window text
			int text_len = GetWindowTextLengthW(handle);
			std::wstring text(text_len, L'\0');
			text_len = GetWindowTextW(handle, text.data(), text_len + 1);
			text.resize(text_len);

			//Get the class name
			//No way to measure the class name's length before allocation, but its is limitted to max 256 characters
			std::wstring class_name(256, L'\0');
			text_len = GetClassNameW(handle, class_name.data(), 257);
			class_name.resize(text_len);

			//output result
			std::wostringstream buf{};
			buf << prefix << handle << L'!' << class_name;
			if (!text.empty())
				buf << L" (" << text << L')';

			buf << std::endl;
			OutputDebugStringW(buf.str().c_str());
		}

		static BOOL CALLBACK enum_proc(HWND handle, LPARAM param)
		{
			//Get the previous child window and its depth
			auto &[prev_handle, depth] = *reinterpret_cast<std::pair<HWND, size_t> *>(param);

			//Get the parent of current child
			HWND parent = GetParent(handle);

			//Due to traversal order, parent is closes common node in the hierarchy containing the previous and the 
			//current child. It can be prev_handle or one of its ancestors. To find the depth of parent, we can
			//just travel up from prev_handle up the hierarchy.
			while (depth > 0 && prev_handle != parent)
			{
				--depth;
				prev_handle = GetParent(prev_handle);
				assert(prev_handle);
			}

			//the parent's depth determines the size of our indent prefix
			std::wstring prefix(depth * 2 + 2, L' ');
			prefix[prefix.size() - 2] = L'*';

			//Print the child info
			print_window(handle, prefix.c_str());

			//Store current as previous and increase depth
			prev_handle = handle;
			++depth;

			//Return true to continue the enumeration
			return true;
		}
	};
}